@@ -15,10 +15,15 @@
 # limitations under the License.
 
 
+import logging
+import math
+
 from modules.map.proto import map_pb2
 from modules.map.proto import map_road_pb2
 from modules.map.proto import map_lane_pb2
 
+import imap.global_var as global_var
+
 from imap.lib.opendrive.map import Map
 from imap.lib.proto_utils import (
   write_pb_to_text_file,
@@ -26,7 +31,11 @@ from imap.lib.proto_utils import (
 )
 
 from imap.lib.draw import draw_line, show
-from imap.lib.convex_hull import convex_hull
+from imap.lib.convex_hull import convex_hull, aabb_box
+
+
+# Distance between stop line and pedestrian crossing
+STOP_LINE_DISTANCE = 1.0
 
 
 def to_pb_lane_type(open_drive_type):
@@ -107,6 +116,9 @@ class Opendrive2Apollo(Convertor):
     self.xodr_map = Map()
     self.xodr_map.load(input_file_name)
 
+    # lhd for saving figure
+    self.input_file_name = input_file_name
+
     self.pb_map = map_pb2.Map()
 
     self.output_file_name = self._get_file_name(output_file_name)
@@ -165,7 +177,11 @@ class Opendrive2Apollo(Convertor):
     segment = pb_lane.left_boundary.curve.segment.add()
     for point3d in lane.left_boundary:
       point = segment.line_segment.point.add()
-      point.x, point.y = point3d.x, point3d.y
+      # lhd 2022/12/03 for 3D view
+      if global_var.get_element_value("enable_z_axis"):
+        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
+      else:
+        point.x, point.y = point3d.x, point3d.y
     segment.s = 0
     segment.start_position.x = lane.left_boundary[0].x
     segment.start_position.y = lane.left_boundary[0].y
@@ -181,7 +197,11 @@ class Opendrive2Apollo(Convertor):
     segment = pb_lane.central_curve.segment.add()
     for point3d in lane.center_line:
       point = segment.line_segment.point.add()
-      point.x, point.y = point3d.x, point3d.y
+      # lhd 2022/12/03 for 3D view
+      if global_var.get_element_value("enable_z_axis"):
+        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
+      else:
+        point.x, point.y = point3d.x, point3d.y
     segment.s = 0
     segment.start_position.x = lane.center_line[0].x
     segment.start_position.y = lane.center_line[0].y
@@ -192,7 +212,11 @@ class Opendrive2Apollo(Convertor):
     segment = pb_lane.right_boundary.curve.segment.add()
     for point3d in lane.right_boundary:
       point = segment.line_segment.point.add()
-      point.x, point.y = point3d.x, point3d.y
+      # lhd 2022/12/03 for 3D view
+      if global_var.get_element_value("enable_z_axis"):
+        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
+      else:
+        point.x, point.y = point3d.x, point3d.y
     segment.s = 0
     segment.start_position.x = lane.right_boundary[0].x
     segment.start_position.y = lane.right_boundary[0].y
@@ -370,7 +394,11 @@ class Opendrive2Apollo(Convertor):
     segment = pb_boundary_edge.curve.segment.add()
     for point3d in boundary:
       point = segment.line_segment.point.add()
-      point.x, point.y = point3d.x, point3d.y
+      # lhd 2022/12/03 for 3D view
+      if global_var.get_element_value("enable_z_axis"):
+        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
+      else:
+        point.x, point.y = point3d.x, point3d.y
     segment.s = 0
     segment.start_position.x = boundary[0].x
     segment.start_position.y = boundary[0].y
@@ -410,11 +438,75 @@ class Opendrive2Apollo(Convertor):
         if pb_lane is not None:
           pb_road_section.lane_id.add().id = pb_lane.id.id
 
+      pb_last_right_section = []
       for lane in lane_section.right:
         pb_lane = self.create_lane(xodr_road, lane_section, idx, lane)
         if pb_lane is not None:
           pb_road_section.lane_id.add().id = pb_lane.id.id
+          # add right lane
+          pb_last_right_section.append(pb_lane)
+    return pb_last_right_section
 
+  def construct_signal_overlap(self, pb_lane, pb_signal):
+    # lane_overlap_info
+    pb_overlap = self.pb_map.overlap.add()
+    pb_overlap.id.id = "{}_{}".format(pb_lane.id.id, pb_signal.id.id)
+    # lane_overlap_info
+    pb_object = pb_overlap.object.add()
+    pb_object.id.id = pb_lane.id.id
+    # todo(zero): need to complete
+    # pb_object.lane_overlap_info.start_s
+    # pb_object.lane_overlap_info.end_s
+
+    # signal_overlap_info
+    pb_object = pb_overlap.object.add()
+    pb_object.id.id = pb_signal.id.id
+
+    # add overlap to lane and signal
+    pb_lane_overlap_id = pb_lane.overlap_id.add()
+    pb_lane_overlap_id.id = pb_overlap.id.id
+
+    pb_signal_overlap_id = pb_signal.overlap_id.add()
+    pb_signal_overlap_id.id = pb_overlap.id.id
+
+  def _construct_signal_stopline(self, last_section_lanes, pb_signal):
+    if len(last_section_lanes) == 1:
+      pb_left_lane = pb_right_lane = last_section_lanes[0]
+    else:
+      pb_left_lane, pb_right_lane = last_section_lanes[0], last_section_lanes[-1]
+    left_segment = pb_left_lane.left_boundary.curve.segment[-1].line_segment
+    right_segment = pb_right_lane.right_boundary.curve.segment[-1].line_segment
+
+    sampling_length = global_var.get_element_value("sampling_length")
+    index = math.ceil(STOP_LINE_DISTANCE/sampling_length)
+    pb_stop_line = pb_signal.stop_line.add()
+    pb_segment = pb_stop_line.segment.add()
+
+    if len(left_segment.point) < index:
+      point = pb_segment.line_segment.point.add()
+      point.CopyFrom(left_segment.point[-1])
+      point = pb_segment.line_segment.point.add()
+      point.CopyFrom(right_segment.point[-1])
+    else:
+      point = pb_segment.line_segment.point.add()
+      point.CopyFrom(left_segment.point[-index])
+      point = pb_segment.line_segment.point.add()
+      point.CopyFrom(right_segment.point[-index])
+
+  def convert_signal(self, xodr_road, pb_last_right_section):
+    if not pb_last_right_section:
+      return
+
+    for signal in xodr_road.signals.signals:
+      # todo(zero): ref 2.3.7. Specific examples of OpenDRIVE objects
+      # need to check 'signal.type'
+      pb_signal = self.pb_map.signal.add()
+      # add road_id to avoid duplication
+      pb_signal.id.id = "signal_{}_{}".format(xodr_road.road_id, signal.id)
+      # todo(zero): needs to be completed(boundary\subsignal\type)
+      self._construct_signal_stopline(pb_last_right_section, pb_signal)
+      for pb_lane in pb_last_right_section:
+        self.construct_signal_overlap(pb_lane, pb_signal)
 
   def convert_roads(self):
     for _, xodr_road in self.xodr_map.roads.items():
@@ -430,49 +522,64 @@ class Opendrive2Apollo(Convertor):
       xodr_road.generate_reference_line()
       xodr_road.add_offset_to_reference_line()
       # Todo(zero):
-      draw_line(xodr_road.reference_line, 'r')
+      draw_line(xodr_road.reference_line, 'r', \
+        reference_line = True, label = "reference line " + str(pb_road.id.id))
 
       xodr_road.process_lanes()
 
-      self.convert_lane(xodr_road, pb_road)
+      pb_last_right_section = self.convert_lane(xodr_road, pb_road)
+      # Todo(zero): need to complete signal
+      # self.convert_signal(xodr_road, pb_last_right_section)
 
+  def _is_valid_junction(self, xodr_junction):
+    connecting_roads = set()
+    incoming_roads = set()
+    for connection in xodr_junction.connections:
+      connecting_roads.add(connection.connecting_road)
+      incoming_roads.add(connection.incoming_road)
+
+    return len(connecting_roads) != 1 and len(incoming_roads) != 1
 
   def construct_junction_polygon(self, xodr_junction):
+    if not self._is_valid_junction(xodr_junction):
+      return []
+
     points = []
     for road, relation in xodr_junction.connected_roads:
       start, end = road.get_cross_section(relation)
       points.append([start.x, start.y])
       points.append([end.x, end.y])
 
-    # order
-    return convex_hull(points)
+    # when point <= 4 convex_hull will not fully covered, so we change to aabb_box
+    if len(points) <= 4:
+      return aabb_box(points)
+    else:
+      return convex_hull(points)
 
 
   def convert_junctions(self):
     for _, xodr_junction in self.xodr_map.junctions.items():
+      polygon = self.construct_junction_polygon(xodr_junction)
+      if len(polygon) < 3:
+        logging.warning(
+          "junction {} polygon size < 3.".format(xodr_junction.junction_id))
+        continue
+
       pb_junction = self.pb_map.junction.add()
       pb_junction.id.id = xodr_junction.junction_id
-      polygon = self.construct_junction_polygon(xodr_junction)
       for x, y in polygon:
         pb_point = pb_junction.polygon.point.add()
         pb_point.x, pb_point.y, pb_point.z = x, y, 0
 
-
-  def convert_overlap(self):
-    # lane_overlap_info
-    pass
-
-
   def convert(self):
     self.convert_header()
     # Don't change the order. "convert_roads" must before "convert_junctions"
     self.convert_roads()
     self.convert_junctions()
-    self.convert_overlap()
 
     # Todo(zero): display xodr map
     if self.output_file_name is None:
-      show()
+      show(need_save=global_var.get_element_value("need_save_figure"), path=self.input_file_name.replace(".xodr", ".png"))
 
 
   def save_map(self):
